/* Revision History
 *
 * 14-MAR-2013 (EC):
 *   - Added support for interpr-specific types Ulong and Long so that
 *     we can produce arrays that describe dimensions for certain dal
 *     calls (i.e. addColumn). Ulong is the most complete
 *     implementation, and more-or-less matches the implementation of
 *     int.
 *
 * $Id: token.cc,v 1.20 2013/03/14 14:24:02 echapin Exp $
 *
 */


/**************************************************************************/
/*                                                                        */
/*  FILENAME                                                              */
/*  ========                                                              */
/*  token.cc                                                              */
/*                                                                        */
/*                                                                        */
/*  SUBSYSTEM                                                             */
/*  =========                                                             */
/*  XMM SAS - UTILITIES                                                   */
/*                                                                        */
/*                                                                        */
/*  TITLE                                                                 */
/*  =====                                                                 */
/*  Implementation file for TokenT type.                                  */
/*                                                                        */
/*                                                                        */
/*  AUTHORS                                                               */
/*  =======                                                               */
/*  Mark Thomas                                                           */
/*  CARA Software & Services Ltd                                          */
/*                                                                        */
/*                                                                        */
/*  CONSTRAINTS                                                           */
/*  ===========                                                           */
/*                                                                        */
/*                                                                        */
/*  DESCRIPTION                                                           */
/*  ===========                                                           */
/*  Implementation file for TokenT type.                                  */
/*                                                                        */
/*                                                                        */
/**************************************************************************/

#include "token.h"

#include <cstdlib>
#include <cstring>
#include <cstdarg>
#include <list>

#include "binary_tree.h"
#include "Tracer.h"
#include "stringutil.h"


static void
invalidConversion()
{
	PROC
	cerr << "The requested Token data conversion is not possible" << endl;
	exit( 1 );
	ENDPROC
}


static void
invalidAddress()
{
	PROC
	cerr << "Invalid address" << endl;
	exit( 1 );
	ENDPROC
}


static void
invalidOperation( string op )
{
	PROC
	cerr << "The operation '" << op << "' is not valid" << endl;
	exit( 1 );
	ENDPROC
}


class TokenValue
{
	public :
		virtual ~TokenValue() {}
		virtual int asInt() const
		{ PROC invalidConversion(); RETURN 0; }
		virtual int asShort() const
		{ PROC invalidConversion(); RETURN 0; }
		virtual bool asBool() const
		{ PROC invalidConversion(); RETURN true; }
		virtual unsigned int asUInt() const
		{ PROC invalidConversion(); RETURN 0; }

                // new from EC
  		virtual unsigned long asUlong() const
		{ PROC invalidConversion(); RETURN 0; }
  		virtual long asLong() const
		{ PROC invalidConversion(); RETURN 0; }

		virtual double asReal() const
		{ PROC invalidConversion(); RETURN 0.0; }
		virtual float asSingle() const
		{ PROC invalidConversion(); RETURN 0.0; }
		virtual char asCharacter() const
		{ PROC invalidConversion(); RETURN 0; }
		virtual string asString() const
		{ PROC invalidConversion(); RETURN ""; }
		virtual const void * asPointer() const
		{ PROC invalidConversion(); RETURN static_cast<const void *>( 0 ); }
		virtual Functor asFunctor() const
		{ PROC invalidConversion(); RETURN Functor(Functor::C, 0, PointerDescriptor( Token::Integer, 0, 0 )); }
		virtual PointerDescriptor asPointerDescriptor() const
		{ PROC invalidConversion(); RETURN PointerDescriptor(Token::Integer,0, 0); }
		virtual const Vec<Token> & asVector() const
		{ PROC static Vec<Token> dummy; invalidConversion(); RETURN dummy; }
		virtual const vector<Token::BTree> & asBTreeVector() const
		{ PROC static vector<Token::BTree> dummy; invalidConversion(); RETURN dummy; }
		virtual const Token::BTree & asBTree() const
		{ PROC static Token::BTree dummy; invalidConversion(); RETURN dummy; }
		virtual Token asToken() const
		{ PROC invalidConversion(); RETURN Token(); }
		virtual void setAddressOfValue( const int * address )
		{ PROC address = address; invalidAddress(); ENDPROC }
		virtual void setAddressOfValue( const unsigned int * address )
		{ PROC address = address; invalidAddress(); ENDPROC }

                // new from EC
		virtual void setAddressOfValue( const long * address )
		{ PROC address = address; invalidAddress(); ENDPROC }
		virtual void setAddressOfValue( const unsigned long * address )
		{ PROC address = address; invalidAddress(); ENDPROC }

		virtual void setAddressOfValue( const double * address )
		{ PROC address = address; invalidAddress(); ENDPROC }
		virtual void setAddressOfValue( const string * address )
		{ PROC address = address; invalidAddress(); ENDPROC }
		virtual const void * addressOfValue() const
		{ PROC invalidAddress(); RETURN static_cast<void *>( 0 ); }
		virtual bool byReference() const
		{ PROC RETURN false; }
		virtual Token operator=( const Token & token )
		{ PROC invalidOperation("="); cerr << sizeof( token ); RETURN 0; }
		virtual Token operator+( const TokenValue & value ) const
		{ PROC invalidOperation("+"); cerr << sizeof( value ); RETURN 0; }

		virtual Token operator+( const string & value ) const
		{ PROC invalidOperation("+"); cerr << sizeof( value ); RETURN 0; }

		virtual Token operator+( int value ) const
		{ PROC invalidOperation("+"); cerr << sizeof( value ); RETURN 0; }

		virtual Token operator+( double value ) const
		{ PROC invalidOperation("+"); cerr << sizeof( value ); RETURN 0; }

		virtual Token operator-( const TokenValue & value ) const
		{ PROC invalidOperation("-"); cerr << sizeof( value ); RETURN 0; }

		virtual Token minus( int value ) const
		{ PROC invalidOperation("-"); cerr << value; RETURN 0; }

		virtual Token minus(double value ) const
		{ PROC invalidOperation("-"); cerr << value; RETURN 0; }

		virtual Token operator-() const
		{ PROC invalidOperation("-"); RETURN 0; }

		virtual Token operator*( const TokenValue & value ) const
		{ PROC invalidOperation("*"); cerr << sizeof( value ); RETURN 0; }

		virtual Token operator*( int value ) const
		{ PROC invalidOperation("*"); cerr << sizeof( value ); RETURN 0; }

		virtual Token operator*( double value ) const
		{ PROC invalidOperation("*"); cerr << sizeof( value ); RETURN 0; }

		virtual Token operator/( const TokenValue & value ) const
		{ PROC invalidOperation("/"); cerr << sizeof( value ); RETURN 0; }

		virtual Token div( int value ) const
		{ PROC invalidOperation("/"); cerr << value; RETURN 0; }

		virtual Token div(double value ) const
		{ PROC invalidOperation("/"); cerr << value; RETURN 0; }

		virtual bool operator<( const TokenValue & value ) const
		{ PROC invalidOperation("<"); cerr << sizeof( value ); RETURN 0; }

		virtual bool operator<( int value ) const
		{ PROC invalidOperation("<"); cerr << sizeof( value ); RETURN 0; }

		virtual bool operator<( double value ) const
		{ PROC invalidOperation("<"); cerr << sizeof( value ); RETURN 0; }

		virtual bool lss( int value ) const
		{ PROC invalidOperation("<"); cerr << sizeof( value ); RETURN 0; }

		virtual bool operator>( const TokenValue & value ) const
		{ PROC invalidOperation(">"); cerr << sizeof( value ); RETURN 0; }

		virtual bool operator>( int value ) const
		{ PROC invalidOperation(">"); cerr << sizeof( value ); RETURN 0; }

		virtual bool operator>( double value ) const
		{ PROC invalidOperation(">"); cerr << sizeof( value ); RETURN 0; }

		virtual bool gtr( int value ) const
		{ PROC invalidOperation(">"); cerr << sizeof( value ); RETURN 0; }

		virtual bool operator>=( const TokenValue & value ) const
		{ PROC invalidOperation(">="); cerr << sizeof( value ); RETURN 0; }

		virtual bool operator>=( int value ) const
		{ PROC invalidOperation(">="); cerr << sizeof( value ); RETURN 0; }

		virtual bool operator>=( double value ) const
		{ PROC invalidOperation(">="); cerr << sizeof( value ); RETURN 0; }

		virtual bool gtreq( int value ) const
		{ PROC invalidOperation(">="); cerr << sizeof( value ); RETURN 0; }

		virtual bool operator<=( const TokenValue & value ) const
		{ PROC invalidOperation("<="); cerr << sizeof( value ); RETURN 0; }

		virtual bool operator<=( int value ) const
		{ PROC invalidOperation("<="); cerr << sizeof( value ); RETURN 0; }

		virtual bool operator<=( double value ) const
		{ PROC invalidOperation("<="); cerr << sizeof( value ); RETURN 0; }

		virtual bool lsseq( int value ) const
		{ PROC invalidOperation("<="); cerr << sizeof( value ); RETURN 0; }

		virtual bool operator==( const TokenValue & value ) const
		{ PROC invalidOperation("=="); cerr << sizeof( value ); RETURN 0; }

		virtual bool operator==( int value ) const
		{ PROC invalidOperation("=="); cerr << sizeof( value ); RETURN 0; }

		virtual bool operator==( double value ) const
		{ PROC invalidOperation("=="); cerr << sizeof( value ); RETURN 0; }

		virtual bool eqeq( int value ) const
		{ PROC invalidOperation("=="); cerr << sizeof( value ); RETURN 0; }


                // new from EC
		virtual Token operator+( long value ) const
		{ PROC invalidOperation("+"); cerr << sizeof( value ); RETURN 0; }
		virtual Token minus( long value ) const
		{ PROC invalidOperation("-"); cerr << sizeof( value ); RETURN 0; }
		virtual Token operator*( long value ) const
		{ PROC invalidOperation("*"); cerr << sizeof( value ); RETURN 0; }
		virtual Token div( long value ) const
		{ PROC invalidOperation("/"); cerr << sizeof( value ); RETURN 0; }
		virtual bool operator<( long value ) const
		{ PROC invalidOperation("<"); cerr << sizeof( value ); RETURN 0; }
		virtual bool lss( long value ) const
		{ PROC invalidOperation("<"); cerr << sizeof( value ); RETURN 0; }
		virtual bool operator>( long value ) const
		{ PROC invalidOperation(">"); cerr << sizeof( value ); RETURN 0; }
		virtual bool gtr( long value ) const
		{ PROC invalidOperation(">"); cerr << sizeof( value ); RETURN 0; }
		virtual bool operator>=( long value ) const
		{ PROC invalidOperation(">="); cerr << sizeof( value ); RETURN 0; }
		virtual bool gtreq( long value ) const
		{ PROC invalidOperation(">="); cerr << sizeof( value ); RETURN 0; }
		virtual bool operator<=( long value ) const
		{ PROC invalidOperation("<="); cerr << sizeof( value ); RETURN 0; }
		virtual bool lsseq( long value ) const
		{ PROC invalidOperation("<="); cerr << sizeof( value ); RETURN 0; }
		virtual bool operator==( long value ) const
		{ PROC invalidOperation("=="); cerr << sizeof( value ); RETURN 0; }
		virtual bool eqeq( long value ) const
		{ PROC invalidOperation("=="); cerr << sizeof( value ); RETURN 0; }

		virtual void printOn( ostream & ) const
		{ PROC ENDPROC }
};


class IntTokenValue : public virtual TokenValue
{
	public :
		IntTokenValue( int value ) : _value( value ), _addressOfValue( &_value ) { PROC ENDPROC }
		//IntTokenValue( int value ) : _value( value ), _addressOfValue( 0 ) { PROC ENDPROC }
		IntTokenValue( const int * value ) : _value( *value ), _addressOfValue( value ) { PROC ENDPROC }
		~IntTokenValue() {}

		//int asInt() const { PROC RETURN _value; }

		int asInt() const { PROC RETURN *_addressOfValue; }

		char asCharacter() const { PROC RETURN _value; }

		const void * asPointer() const { PROC RETURN (const void *)_value; }

		unsigned int asUInt() const { PROC RETURN _value; }


                // new from EC
                long asLong() const { PROC RETURN _value; }
                unsigned long asUlong() const { PROC RETURN _value; }


		double asReal() const { PROC RETURN _value; }

		string asString() const { PROC RETURN toString( _value ); }

		bool asBool() const { PROC RETURN bool( _value ); }

		void setAddressOfValue( const int * address ) { PROC _addressOfValue = address; ENDPROC }
		const void * addressOfValue() const { PROC RETURN _addressOfValue; }

		Token operator=( const Token & token )
		{ PROC *const_cast<int *>( _addressOfValue ) = token.asInt(); RETURN _value; }

		bool byReference() const { PROC RETURN _addressOfValue != &_value; }

		virtual Token operator+( const TokenValue & value ) const
		{ PROC RETURN value + _value; }

		virtual Token operator+( int value ) const
		{ PROC RETURN _value + value; }

		virtual Token operator+( double value ) const
		{ PROC RETURN _value + value; }

		virtual Token operator+( const string & value ) const
		{ PROC RETURN value + toString( _value ); }

		virtual Token operator-( const TokenValue & value ) const
		{ PROC RETURN value.minus( _value ); }

		Token minus( int value ) const
		{ PROC RETURN value - _value; }

		Token minus( double value ) const
		{ PROC RETURN value - _value; }

		virtual Token operator-() const
		{ PROC RETURN -_value; }

		virtual Token operator*( const TokenValue & value ) const
		{ PROC RETURN value * _value; }

		virtual Token operator*( int value ) const
		{ PROC RETURN _value * value; }

		virtual Token operator*( double value ) const
		{ PROC RETURN _value * value; }

		Token operator/( const TokenValue & value ) const
		{ PROC RETURN value.div(_value); }

		Token div( int value ) const
		{ PROC RETURN value / _value; }

		Token div( double value ) const
		{ PROC RETURN value / _value; }

		bool operator<( const TokenValue & value ) const
		{ PROC RETURN value.lss( _value ); }

		bool lss( int value ) const
		{ PROC RETURN value < _value; }

		bool operator<=( const TokenValue & value ) const
		{ PROC RETURN value.lsseq( _value ); }

		bool lsseq( int value ) const
		{ PROC RETURN value <= _value; }

		bool operator>( const TokenValue & value ) const
		{ PROC RETURN value.gtr( _value ); }

		bool gtr( int value ) const
		{ PROC RETURN value > _value; }

		bool operator>=( const TokenValue & value ) const
		{ PROC RETURN value.gtreq( _value ); }

		bool gtreq( int value ) const
		{ PROC RETURN value >= _value; }

		bool operator==( const TokenValue & value ) const
		{ PROC RETURN value.eqeq( _value ); }

		bool eqeq( int value ) const
		{ PROC RETURN value == _value; }

		void printOn( ostream & os ) const
		{ PROC os << "type: Int" << " address: " << _addressOfValue << " value: " << _value << endl; ENDPROC }

	private :
		int 		_value;
		const int *	_addressOfValue;
};

// new from EC
class LongTokenValue : public virtual TokenValue
{
	public :
		LongTokenValue( long value ) : _value( value ), _addressOfValue( &_value ) { PROC ENDPROC }
		//LongtokenValue( long value ) : _value( value ), _addressOfValue( 0 ) { PROC ENDPROC }
		LongTokenValue( const long * value ) : _value( *value ), _addressOfValue( value ) { PROC ENDPROC }
		~LongTokenValue() {}

		//long asLong() const { PROC RETURN _value; }

		long asLong() const { PROC RETURN *_addressOfValue; }

		char asCharacter() const { PROC RETURN _value; }

		const void * asPointer() const { PROC RETURN (const void *)_value; }

		unsigned long asUlong() const { PROC RETURN _value; }

		double asReal() const { PROC RETURN _value; }

		string asString() const { PROC RETURN toString( _value ); }

		bool asBool() const { PROC RETURN bool( _value ); }

		void setAddressOfValue( const long * address ) { PROC _addressOfValue = address; ENDPROC }
		const void * addressOfValue() const { PROC RETURN _addressOfValue; }

		Token operator=( const Token & token )
		{ PROC *const_cast<long *>( _addressOfValue ) = token.asLong(); RETURN _value; }

		bool byReference() const { PROC RETURN _addressOfValue != &_value; }

		virtual Token operator+( const TokenValue & value ) const
		{ PROC RETURN value + _value; }

		virtual Token operator+( long value ) const
		{ PROC RETURN _value + value; }

		virtual Token operator+( double value ) const
		{ PROC RETURN _value + value; }

		virtual Token operator+( const string & value ) const
		{ PROC RETURN value + toString( _value ); }

		virtual Token operator-( const TokenValue & value ) const
		{ PROC RETURN value.minus( _value ); }

		Token minus( long value ) const
		{ PROC RETURN value - _value; }

		Token minus( double value ) const
		{ PROC RETURN value - _value; }

		virtual Token operator-() const
		{ PROC RETURN -_value; }

		virtual Token operator*( const TokenValue & value ) const
		{ PROC RETURN value * _value; }

		virtual Token operator*( long value ) const
		{ PROC RETURN _value * value; }

		virtual Token operator*( double value ) const
		{ PROC RETURN _value * value; }

		Token operator/( const TokenValue & value ) const
		{ PROC RETURN value.div(_value); }

		Token div( long value ) const
		{ PROC RETURN value / _value; }

		Token div( double value ) const
		{ PROC RETURN value / _value; }

		bool operator<( const TokenValue & value ) const
		{ PROC RETURN value.lss( _value ); }

		bool lss( long value ) const
		{ PROC RETURN value < _value; }

		bool operator<=( const TokenValue & value ) const
		{ PROC RETURN value.lsseq( _value ); }

		bool lsseq( long value ) const
		{ PROC RETURN value <= _value; }

		bool operator>( const TokenValue & value ) const
		{ PROC RETURN value.gtr( _value ); }

		bool gtr( long value ) const
		{ PROC RETURN value > _value; }

		bool operator>=( const TokenValue & value ) const
		{ PROC RETURN value.gtreq( _value ); }

		bool gtreq( long value ) const
		{ PROC RETURN value >= _value; }

		bool operator==( const TokenValue & value ) const
		{ PROC RETURN value.eqeq( _value ); }

		bool eqeq( long value ) const
		{ PROC RETURN value == _value; }

		void printOn( ostream & os ) const
		{ PROC os << "type: Long" << " address: " << _addressOfValue << " value: " << _value << endl; ENDPROC }

	private :
		long 		_value;
		const long *	_addressOfValue;
};

class ShortTokenValue : public virtual TokenValue
{
	public :
		ShortTokenValue( short value ) : _value( value ), _addressOfValue( &_value ) { PROC ENDPROC }
		ShortTokenValue( const short * value ) : _value( *value ), _addressOfValue( value ) { PROC ENDPROC }
		~ShortTokenValue() {}

		int asShort() const { PROC RETURN *_addressOfValue; }
		int asInt() const { PROC RETURN *_addressOfValue; }
		char asCharacter() const { PROC RETURN _value; }
		unsigned int asUInt() const { PROC RETURN _value; }
		double asReal() const { PROC RETURN _value; }
		string asString() const { PROC RETURN toString( _value ); }
		void setAddressOfValue( const short * address ) { PROC _addressOfValue = address; ENDPROC }
		const void * addressOfValue() const { PROC RETURN _addressOfValue; }
		Token operator=( const Token & token )
		{ PROC *const_cast<short *>( _addressOfValue ) = token.asShort(); RETURN _value; }
		bool byReference() const { PROC RETURN _addressOfValue != &_value; }
		void printOn( ostream & os ) const
		{ PROC os << "type: Short" << " address: " << _addressOfValue << " value: " << _value << endl; ENDPROC }

	private :
		short		_value;
		const short *	_addressOfValue;
};


class BoolTokenValue : public virtual TokenValue
{
	public :
		BoolTokenValue( bool value ) : _value( value ), _addressOfValue( &_value ) { PROC ENDPROC }
		BoolTokenValue( const bool * value ) : _value( *value ), _addressOfValue( value ) { PROC ENDPROC }
		~BoolTokenValue() {}
		int asInt() const { PROC RETURN int(_value); }
		bool asBool() const { PROC RETURN *_addressOfValue; }
		char asCharacter() const { PROC RETURN _value; }
		const void * asPointer() const { PROC RETURN (const void *)_value; }
		const void * addressOfValue() const { PROC RETURN _addressOfValue; }
		Token operator=( const Token & token )
		{ PROC *const_cast<bool *>( _addressOfValue ) = token.asBool(); RETURN _value; }
		bool byReference() const { PROC RETURN _addressOfValue != &_value; }
		void printOn( ostream & os ) const
		{ PROC os << "type: Bool" << " address: " << _addressOfValue << " value: " << _value << endl; ENDPROC }

	private :
		bool		_value;
		const bool *	_addressOfValue;
};


class UIntTokenValue : public virtual TokenValue
{
	public :
		UIntTokenValue( unsigned int value ) : _value( value ), _addressOfValue( &_value ) { PROC ENDPROC }
		UIntTokenValue( const unsigned int * value ) : _value( *value ), _addressOfValue( value ) { PROC ENDPROC }
		~UIntTokenValue() {}
		int asInt() const { PROC RETURN _value; }
		unsigned int asUInt() const { PROC RETURN _value; }

                // new from EC
                long asLong() const { PROC RETURN _value; }
                unsigned long asUlong() const { PROC RETURN _value; }

		double asReal() const { PROC RETURN _value; }
		string asString() const { PROC RETURN toString( _value ); }
		void setAddressOfValue( const unsigned int * address ) { PROC _addressOfValue = address; ENDPROC }
		const void * addressOfValue() const { PROC RETURN _addressOfValue; }
		void printOn( ostream & os ) const
		{ PROC os << "type: UInt" << " address: " << _addressOfValue << " value: " << _value << endl; ENDPROC }

	private :
		unsigned int 		_value;
		const unsigned int *	_addressOfValue;
};

// new from EC
class UlongTokenValue : public virtual TokenValue
{
	public :
		UlongTokenValue( unsigned long value ) : _value( value ), _addressOfValue( &_value ) { PROC ENDPROC }
		//UlongTokenValue( unsigned long value ) : _value( value ), _addressOfValue( 0 ) { PROC ENDPROC }
		UlongTokenValue( const unsigned long * value ) : _value( *value ), _addressOfValue( value ) { PROC ENDPROC }
		~UlongTokenValue() {}

		//unsigned long asUlong() const { PROC RETURN _value; }

		unsigned long asUlong() const { PROC RETURN *_addressOfValue; }

		char asCharacter() const { PROC RETURN _value; }

		const void * asPointer() const { PROC RETURN (const void *)_value; }

                // new from EC
                long asLong() const { PROC RETURN _value; }

		double asReal() const { PROC RETURN _value; }

		string asString() const { PROC RETURN toString( _value ); }

		bool asBool() const { PROC RETURN bool( _value ); }

		void setAddressOfValue( const unsigned long * address ) { PROC _addressOfValue = address; ENDPROC }
		const void * addressOfValue() const { PROC RETURN _addressOfValue; }

		Token operator=( const Token & token )
		{ PROC *const_cast<unsigned long *>( _addressOfValue ) = token.asUlong(); RETURN _value; }

		bool byReference() const { PROC RETURN _addressOfValue != &_value; }

		virtual Token operator+( unsigned long value ) const
		{ PROC RETURN _value + value; }

		virtual Token operator+( double value ) const
		{ PROC RETURN _value + value; }

		virtual Token operator+( const string & value ) const
		{ PROC RETURN value + toString( _value ); }

		Token minus( unsigned long value ) const
		{ PROC RETURN value - _value; }

		Token minus( double value ) const
		{ PROC RETURN value - _value; }

		virtual Token operator-() const
		{ PROC RETURN -_value; }

		virtual Token operator*( unsigned long value ) const
		{ PROC RETURN _value * value; }

		virtual Token operator*( double value ) const
		{ PROC RETURN _value * value; }

		Token div( unsigned long value ) const
		{ PROC RETURN value / _value; }

		Token div( double value ) const
		{ PROC RETURN value / _value; }

		bool lss( unsigned long value ) const
		{ PROC RETURN value < _value; }

		bool lsseq( unsigned long value ) const
		{ PROC RETURN value <= _value; }

		bool gtr( unsigned long value ) const
		{ PROC RETURN value > _value; }

		bool gtreq( unsigned long value ) const
		{ PROC RETURN value >= _value; }

		bool eqeq( unsigned long value ) const
		{ PROC RETURN value == _value; }

		void printOn( ostream & os ) const
		{ PROC os << "type: Ulong" << " address: " << _addressOfValue << " value: " << _value << endl; ENDPROC }

	private :
		unsigned long 		_value;
		const unsigned long *	_addressOfValue;
};


class RealTokenValue : public virtual TokenValue
{
	public :
		RealTokenValue( double value ) : _value( value ), _addressOfValue( &_value ) { PROC ENDPROC }
		RealTokenValue( const double * value ) : _value( *value ), _addressOfValue( value ) { PROC ENDPROC }
		~RealTokenValue() {}
		int asInt() const { PROC RETURN int(_value); }
		unsigned int asUInt() const { PROC RETURN (unsigned int)(_value); }
		double asReal() const { PROC RETURN _value; }
		string asString() const { PROC RETURN toString( _value ); }
		void setAddressOfValue( const double * address ) { PROC _addressOfValue = address; ENDPROC }
		const void * addressOfValue() const { PROC RETURN _addressOfValue; }

		Token operator-( const TokenValue & value ) const { PROC RETURN value.minus(_value); }
		Token minus( int value ) const { PROC RETURN value - _value; }
		Token minus( double value ) const { PROC RETURN value - _value; }
		Token operator/( const TokenValue & value ) const { PROC RETURN value.div(_value); }
		Token div( int value ) const { PROC RETURN value / _value; }
		Token div( double value ) const { PROC RETURN value / _value; }
		void printOn( ostream & os ) const
		{ PROC os << "type: Real" << " address: " << _addressOfValue << " value: " << _value << endl; ENDPROC }

	private :
		double 		_value;
		const double *	_addressOfValue;
};


class SingleTokenValue : public virtual TokenValue
{
	public :
		SingleTokenValue( float value ) : _value( value ), _addressOfValue( &_value ) { PROC ENDPROC }
		SingleTokenValue( const float * value ) : _value( *value ), _addressOfValue( value ) { PROC ENDPROC }
		~SingleTokenValue() {}
		int asInt() const { PROC RETURN int(_value); }
		unsigned int asUInt() const { PROC RETURN (unsigned int)(_value); }
		double asReal() const { PROC RETURN _value; }
		float asSingle() const { PROC RETURN _value; }
		string asString() const { PROC RETURN toString( _value ); }
		void setAddressOfValue( const float * address ) { PROC _addressOfValue = address; ENDPROC }
		const void * addressOfValue() const { PROC RETURN _addressOfValue; }
		void printOn( ostream & os ) const
		{ PROC os << "type: Single" << " address: " << _addressOfValue << " value: " << _value << endl; ENDPROC }

	private :
		float		_value;
		const float *	_addressOfValue;
};


class StringTokenValue : public virtual TokenValue
{
	public :
		StringTokenValue( const string & value ) : _value( value ), _addressOfValue( &_value ) { PROC ENDPROC }
		~StringTokenValue() {}
		string asString() const { PROC RETURN _value; }
		const void * asPointer() const { PROC RETURN (const void *)_value.c_str(); }
		void setAddressOfValue( const string * address ) { PROC _addressOfValue = address; ENDPROC }
		const void * addressOfValue() const { PROC RETURN _addressOfValue; }
		virtual Token operator+( const TokenValue & value ) const
		{ PROC RETURN value + _value; }
		virtual Token operator+( const string & value ) const
		{ PROC RETURN value + _value; }
		void printOn( ostream & os ) const
		{ PROC os << "type: String" << " address: " << _addressOfValue << " value: " << _value << endl; ENDPROC }


	private :
		string 		_value;
		const string *	_addressOfValue;
};


class VectorTokenValue : public virtual TokenValue
{
	public :
		VectorTokenValue( const Vec<Token> & value ) : _value( value ), _addressOfValue( &_value ) { PROC ENDPROC }
		~VectorTokenValue() {}
		const Vec<Token> & asVector() const { PROC RETURN _value; }
		void setAddressOfValue( const Vec<Token> * address ) { PROC _addressOfValue = address; ENDPROC }
		const void * addressOfValue() const { PROC RETURN _addressOfValue; }

	private :
		Vec<Token>			_value;
		const Vec<Token> *		_addressOfValue;
};


class BTreeTokenValue : public virtual TokenValue
{
	public :
		BTreeTokenValue( const Token::BTree & value ) : _value( value ), _addressOfValue( &_value ) { PROC ENDPROC }
		~BTreeTokenValue() {}
		const Token::BTree & asBTree() const { PROC RETURN _value; }
		void setAddressOfValue( const Token::BTree * address ) { PROC _addressOfValue = address; ENDPROC }
		const void * addressOfValue() const { PROC RETURN _addressOfValue; }
	private :
		Token::BTree			_value;
		const Token::BTree *		_addressOfValue;
};


class BTreeVectorTokenValue : public virtual TokenValue
{
	public :
		BTreeVectorTokenValue( const vector<Token::BTree> & value ) : _value( value ), _addressOfValue( &_value ) { PROC ENDPROC }
		~BTreeVectorTokenValue() {}
		const vector<Token::BTree> & asBTreeVector() const { PROC RETURN _value; }
		void setAddressOfValue( const vector<Token::BTree> * address ) { PROC _addressOfValue = address; ENDPROC }
		const void * addressOfValue() const { PROC RETURN _addressOfValue; }

	private :
		vector<Token::BTree>		_value;
		const vector<Token::BTree> *	_addressOfValue;
};


class FunctorTokenValue : public virtual TokenValue
{
	public :
		FunctorTokenValue( const Functor & value ) : _value( value ), _addressOfValue( &_value ) { PROC ENDPROC }

		Functor asFunctor() const { PROC RETURN _value; }
		const void * addressOfValue() const { PROC RETURN _addressOfValue; }
		void printOn( ostream & os ) const
		{ PROC os << "type: Functor" << " address: " << _addressOfValue << " value: ";
		_value.printOn( os );
		os << endl; ENDPROC }

	private :
		Functor		_value;
		const Functor *	_addressOfValue;

};


class PointerDescriptorTokenValue : public virtual TokenValue
{
	public :
		PointerDescriptorTokenValue( const PointerDescriptor & value ) : _value( value ), _addressOfValue( &_value ) { PROC ENDPROC }

		PointerDescriptor asPointerDescriptor() const { PROC RETURN _value; }
		const void * asPointer() const { PROC RETURN _value.raw(); }
		const void * addressOfValue() const { PROC RETURN _addressOfValue; }
		void printOn( ostream & os ) const
		{ PROC os << "type: PointerDescriptor" << " address: " << _addressOfValue << " value: ";
		_value.printOn( os );
		os << endl; ENDPROC }

	private :
		PointerDescriptor		_value;
		const PointerDescriptor * 	_addressOfValue;

};


class CharacterTokenValue : public virtual TokenValue
{
	public :
		CharacterTokenValue( char value ) : _value( value ), _addressOfValue( &_value ) { PROC ENDPROC }
		CharacterTokenValue( const char * value ) : _value( *value ), _addressOfValue( value ) { PROC ENDPROC }

		char asCharacter() const { PROC RETURN *_addressOfValue; }
		const void * asPointer() const { PROC RETURN (const void *)_addressOfValue; }
		const void * addressOfValue() const { PROC RETURN _addressOfValue; }
		bool byReference() const { PROC RETURN _addressOfValue != &_value; }
		Token operator=( const Token & token )
		{ PROC *const_cast<char *>( _addressOfValue ) = token.asCharacter(); RETURN _value; }
		void printOn( ostream & os ) const
		{ PROC os << "type: Character" << " address: " << _addressOfValue << " value: " << _value << endl; ENDPROC }
	private :
		char			_value;
		const char * 		_addressOfValue;

};


class PointerTokenValue : public virtual TokenValue
{
	public :
		PointerTokenValue( const void * value ) : _value( value ), _addressOfValue( &_value ) { PROC ENDPROC }

		const void * asPointer() const { PROC RETURN _value; }
		const void * addressOfValue() const { PROC RETURN _addressOfValue; }
		void printOn( ostream & os ) const
		{ PROC os << "type: Pointer" << " address: " << _addressOfValue << " value: " << _value << endl; ENDPROC }

	private :
		const void *		_value;
		const void * const *	_addressOfValue;

};


Token::Token() : _id( Token::Undefined ), _value( 0 ), _address( 0 )
{
	PROC
	ENDPROC
}


Token::Token( Token::IDType id ) : _id( id ), _value( 0 ), _address( 0 )
{
}


Token::Token( int value ) : _id( Token::Integer ), _value( new IntTokenValue( value )), _address( 0 )
{
	PROC
	ENDPROC
}


// New from EC
Token::Token( long value ) : _id( Token::Long ), _value( new LongTokenValue( value )), _address( 0 )
{
	PROC
	ENDPROC
}
Token::Token( unsigned long value ) : _id( Token::UnsignedLong ), _value( new UlongTokenValue( value )), _address( 0 )
{
	PROC
	ENDPROC
}



Token::Token( short value ) : _id( Token::Short ), _value( new ShortTokenValue( value )), _address( 0 )
{
	PROC
	ENDPROC
}


Token::Token( bool value ) : _id( Token::Boolean ), _value( new BoolTokenValue( value )), _address( 0 )
{

	PROC
	ENDPROC

}


Token::Token( const int * addressOfTheTokenValue ) : _id( Token::Integer ), _value( new IntTokenValue( addressOfTheTokenValue )), _address( 0 )
{

	PROC
	ENDPROC
}

// New from EC
Token::Token( const long * addressOfTheTokenValue ) : _id( Token::Long ), _value( new LongTokenValue( addressOfTheTokenValue )), _address( 0 )
{

	PROC
	ENDPROC
}

Token::Token( const unsigned long * addressOfTheTokenValue ) : _id( Token::UnsignedLong ), _value( new UlongTokenValue( addressOfTheTokenValue )), _address( 0 )
{

	PROC
	ENDPROC
}


//Token::Token( const unsigned int * addressOfTheTokenValue ) : _id( Token::UnsignedInteger ), _value( new UIntTokenValue( addressOfTheTokenValue )), _address( 0 )
//{

//	PROC
//	ENDPROC
//}


Token::Token( double value ) : _id( Token::Real ), _value( new RealTokenValue( value )), _address( 0 )
{
	PROC
	ENDPROC
}


Token::Token( float value ) : _id( Token::Single ), _value( new SingleTokenValue( value )), _address( 0 )
{
	PROC
	ENDPROC
}


Token::Token( const double * value ) : _id( Token::Real ), _value( new RealTokenValue( value )), _address( 0 )
{

	PROC
	ENDPROC

}


Token::Token( char value,  Token::IDType id) : _id( id ), _value( new CharacterTokenValue( value )), _address( 0 )
{

	PROC
	ENDPROC

}



Token::Token( const char * value, Token::IDType id) : _id( id ), _value( new CharacterTokenValue( value )), _address( 0 )
{

	PROC
	ENDPROC

}


Token::Token( const Vec<Token> & value ) : _id( Token::Vector ), _value( new VectorTokenValue( value )), _address( 0 )
{
	PROC
	ENDPROC
}


Token::Token( const vector<Token::BTree> & value, IDType id ) : _id( id ), _value( new BTreeVectorTokenValue( value )), _address( 0 )
{
	PROC
	ENDPROC
}


Token::Token( const BTree & value, IDType id ) : _id( id ), _value( new BTreeTokenValue( value )), _address( 0 )
{
	PROC
	ENDPROC
}


Token::Token( const string & value, IDType id) : _id( id ), _rawValue( value ), _value( new StringTokenValue( value )), _address( 0 )
{

	PROC
	ENDPROC

}


Token::Token( const Functor & value ) : _id( Token::Function ), _value( new FunctorTokenValue( value )), _address( 0 )
{
	PROC
	ENDPROC
}


Token::Token( const PointerDescriptor & value ) : _id( Token::Ptr ), _value( new PointerDescriptorTokenValue( value )), _address( 0 )
{
	PROC
	ENDPROC
}


Token::Token( const void * value, Token::IDType id ) : _id( id ), _value( new PointerTokenValue( value )), _address( 0 )
{
	PROC
	ENDPROC
}


Token::Token( const Token & token )
{
	PROC

	// ********************************************
	// Set _value to null, to prevent, a bad delete
	// (in operator=()).
	// ********************************************
	_value = 0;
	_id = Token::Undefined;

	switch( token.id() )
	{
		//case Token::Integer : operator=( token.asInt() ); break;
		//case Token::Integer : _id = Token::Integer; _value = new IntTokenValue( token.asInt() ); break;
		case Token::Integer :
			_id = Token::Integer;
			if( token._value -> byReference() )
			{
				_value = new IntTokenValue( static_cast<const int *>( token._value -> addressOfValue() ));
			}
			else
			{
				_value = new IntTokenValue( token._value -> asInt() );
			} // if
			break;


                // New from EC
		case Token::Long :
			_id = Token::Long;
			if( token._value -> byReference() )
			{
				_value = new LongTokenValue( static_cast<const long *>( token._value -> addressOfValue() ));
			}
			else
			{
				_value = new LongTokenValue( token._value -> asLong() );
			} // if
			break;

		case Token::Short : operator=( token.asShort() ); break;
		case Token::Boolean : operator=( token.asBool() ); break;
		case Token::Real : operator=( token.asReal() ); break;
		case Token::Single : operator=( token.asSingle() ); break;
		case Token::String : // Fall through
		case Token::FreeText : // Fall through
		case Token::Identifier : operator=( token.asString() ); break;
		//case Token::Character : operator=( token.asCharacter() ); break;
		case Token::Character :
		case Token::QuotedCharacter :
			_id = token.id();
			if( token._value -> byReference() )
			{
				_value = new CharacterTokenValue( static_cast<const char *>( token.addressOfValue() ));
			}
			else
			{
				_value = new CharacterTokenValue( token.asCharacter() );
			} // if
			break;
		case Token::TokenPointer : // Fall through
		case Token::Pointer : operator=( token.asPointer() ); break;
		case Token::Function : operator=( token.asFunctor() ); break;
		case Token::Ptr : operator=( token.asPointerDescriptor() ); break;
		case Token::If : // Fall through
		case Token::Loop : // Fall through
		case Token::BinaryTreeVector : operator=( token.asBTreeVector() ); break;
		case Token::Exitif : // Fall through
		case Token::Return : // Fall through
		case Token::Defun : operator=( token.asBTree() ); break;
		case Token::Vector : operator=( token.asVector() ); break;
		default :
			break;
	} // switch

	_id = token._id;
	_address = token._address;
	_rawValue = token._rawValue;
	_debug = token._debug;

	ENDPROC

}


Token &
Token::operator=( const Token & token )
{
	PROC

/*
	Do some checking to see if token is by ref. or by value.
*/

	//if( _id != Token::Undefined  && _value -> byReference() )
	if( _value != 0 && _value -> byReference() )
	{
		*_value = token;
	}
	else
	{

	switch( token.id() )
	{
		//case Token::Integer : operator=( token.asInt() ); break;
		case Token::Integer :
			delete _value;
			_id = Token::Integer;
			if( token._value -> byReference() )
			{
				_value = new IntTokenValue( static_cast<const int *>( token._value -> addressOfValue() ));
			}
			else
			{
				_value = new IntTokenValue( token._value -> asInt() );
			} // if
			break;

                // new from EC
		case Token::Long :
			delete _value;
			_id = Token::Long;
			if( token._value -> byReference() )
			{
				_value = new LongTokenValue( static_cast<const long *>( token._value -> addressOfValue() ));
			}
			else
			{
				_value = new LongTokenValue( token._value -> asLong() );
			} // if
			break;

		case Token::UnsignedLong :
			delete _value;
			_id = Token::UnsignedLong;
			if( token._value -> byReference() )
			{
				_value = new UlongTokenValue( static_cast<const unsigned long *>( token._value -> addressOfValue() ));
			}
			else
			{
				_value = new UlongTokenValue( token._value -> asUlong() );
			} // if
			break;


		case Token::Short : operator=( token.asShort() ); break;
		case Token::Boolean : operator=( token.asBool() ); break;
		case Token::Real : operator=( token.asReal() ); break;
		case Token::Single : operator=( token.asSingle() ); break;
		case Token::String : // Fall through
		case Token::FreeText : // Fall through
		case Token::Identifier : operator=( token.asString() ); break;
		//case Token::Character : operator=( token.asCharacter() ); break;
		case Token::Character :
		case Token::QuotedCharacter :
			delete _value;

			if( token._value -> byReference() )
			{
				_value = new CharacterTokenValue( static_cast<const char *>( token.addressOfValue() ));
			}
			else
			{
				_value = new CharacterTokenValue( token.asCharacter() );
			} // if
			break;
		case Token::TokenPointer : // Fall through
		case Token::Pointer : operator=( token.asPointer() ); break;
		case Token::Function : operator=( token.asFunctor() ); break;
		case Token::Ptr : operator=( token.asPointerDescriptor() ); break;
		case Token::If : // Fall through
		case Token::Loop : // Fall through
		case Token::BinaryTreeVector : operator=( token.asBTreeVector() ); break;
		case Token::Exitif : // Fall through
		case Token::Return : // Fall through
		case Token::Defun : operator=( token.asBTree() ); break;
		case Token::Vector : operator=( token.asVector() ); break;
		default :
			break;
	} // switch
	_id = token._id;
	_address = token._address;
	} // if

	_rawValue = token._rawValue;
	_debug = token._debug;

	RETURN *this;

}


Token&
Token::operator=( Token::IDType id )
{
	PROC
	delete _value;
	_id = id;
	_value = 0;
	RETURN *this;
}


Token&
Token::operator=( int value )
{
	PROC

        if( _id != Token::Integer )
	{
		delete _value;
		_id = Token::Integer;
		_value = new IntTokenValue( value );
	}
	else
	{
                *(int *)(_value -> addressOfValue()) = value;
	} // if


	//if( _id == Token::Undefined )
	//{
	//	_id = Token::Integer;
	//	_value = new IntTokenValue( value );
	//}
	//else
	//{
	//	if( _value -> byReference() )
	//	{
	//		cerr << "by reference not handled" << endl;
	//		exit( 1 );
	//	}
	//	else
	//	{
			//delete _value;
			//_id = Token::Integer;
			//_value = new IntTokenValue( value );
	//	} // if
	//} // if

	RETURN *this;
}


Token&
Token::operator=( short value )
{
	PROC
	delete _value;
	_id = Token::Short;
	_value = new ShortTokenValue( value );
	RETURN *this;
}

// new from EC
Token&
Token::operator=( long value )
{
	PROC
	delete _value;
	_id = Token::Long;
	_value = new LongTokenValue( value );
	RETURN *this;
}
Token&
Token::operator=( unsigned long value )
{
	PROC
	delete _value;
	_id = Token::UnsignedLong;
	_value = new UlongTokenValue( value );
	RETURN *this;
}


Token&
Token::operator=( bool value )
{
	PROC
	delete _value;
	_id = Token::Boolean;
	_value = new BoolTokenValue( value );
	RETURN *this;
}


Token&
Token::operator=( double value )
{
	PROC
	delete _value;
	_id = Token::Real;
	_value = new RealTokenValue( value );
	RETURN *this;
}


Token&
Token::operator=( float value )
{
	PROC
	delete _value;
	_id = Token::Single;
	_value = new SingleTokenValue( value );
	RETURN *this;
}


Token&
Token::operator=( const string & value )
{
	PROC
	delete _value;
	_id = Token::String;
	_value = new StringTokenValue( value );
	RETURN *this;
}


Token&
Token::operator=( char	value )
{
	PROC
	delete _value;
	_id = Token::Character;
	_value = new CharacterTokenValue( value );
	RETURN *this;
}


Token&
Token::operator=( const Vec<Token> & value )
{
	PROC
	delete _value;
	_id = Token::Vector;
	_value = new VectorTokenValue( value );
	RETURN *this;
}


Token&
Token::operator=( const vector<Token::BTree> & value )
{
	PROC
	delete _value;
	_id = Token::BinaryTreeVector;
	_value = new BTreeVectorTokenValue( value );
	RETURN *this;
}


Token&
Token::operator=( const Token::BTree & value )
{
	PROC
	delete _value;
	_id = Token::Defun;
	_value = new BTreeTokenValue( value );
	RETURN *this;
}


Token&
Token::operator=( const Functor & value )
{
	PROC
	delete _value;
	_id = Token::Function;
	_value = new FunctorTokenValue( value );
	RETURN *this;
}


Token&
Token::operator=( const PointerDescriptor & value )
{
	PROC
	delete _value;
	_id = Token::Ptr ;
	_value = new PointerDescriptorTokenValue( value );
	RETURN *this;
}


Token&
Token::operator=( const void * value )
{
	PROC
	delete _value;
	_id = Token::Pointer;
	_value = new PointerTokenValue( value );
	RETURN *this;
}


Token::~Token()
{
	PROC

	// ***************************
	// Free space for token value.
	// ***************************
	delete _value;

	ENDPROC
}


Token::IDType
Token::id() const
{
	PROC
	RETURN _id;

}


void
Token::setAddressOfValue( const int * toThisAddress )
{
	PROC
	_value -> setAddressOfValue( toThisAddress );
	ENDPROC
}

void
Token::setAddressOfValue( const unsigned int * toThisAddress )
{
	PROC
	_value -> setAddressOfValue( toThisAddress );
	ENDPROC
}


// new from EC
void
Token::setAddressOfValue( const long * toThisAddress )
{
	PROC
	_value -> setAddressOfValue( toThisAddress );
	ENDPROC
}
void
Token::setAddressOfValue( const unsigned long * toThisAddress )
{
	PROC
	_value -> setAddressOfValue( toThisAddress );
	ENDPROC
}


void
Token::setAddressOfValue( const double * toThisAddress )
{
	PROC
	_value -> setAddressOfValue( toThisAddress );
	ENDPROC
}

void
Token::setAddressOfValue( const string * toThisAddress )
{
	PROC
	_value -> setAddressOfValue( toThisAddress );
	ENDPROC
}


const void *
Token::addressOfValue() const
{
	PROC
	RETURN _value-> addressOfValue();

}


const Token *
Token::addressOf() const
{
	PROC
	RETURN _address;
 
}


Token *
Token::addressOf()
{
	PROC
	RETURN const_cast<Token *>( _address );
 
}


void
Token::setAddress( const Token * toThis )
{
	PROC
	_address = toThis;
	ENDPROC

}


void
Token::set(
	const Token &	usingValueOfThisToken )
{

	PROC

	delete _value;

	switch( usingValueOfThisToken.id() )
	{
		case Token::Integer :
			_value = new IntTokenValue( usingValueOfThisToken.asInt() );
			break;
		case Token::UnsignedInteger :
			_value = new UIntTokenValue( usingValueOfThisToken.asUInt() );
			break;

                // new from EC
		case Token::Long :
			_value = new LongTokenValue( usingValueOfThisToken.asLong() );
			break;
		case Token::UnsignedLong :
			_value = new UlongTokenValue( usingValueOfThisToken.asUlong() );
			break;


		case Token::Real :
			_value = new RealTokenValue( usingValueOfThisToken.asReal() );
			break;
		case Token::String :
			_value = new StringTokenValue( usingValueOfThisToken.asString() );
			break;
		default :
			break;
	} // switch

	ENDPROC

}


int 
Token::asInt() const
{
	PROC
	RETURN _value -> asInt();

}

// new from EC
long
Token::asLong() const
{
	PROC
	RETURN _value -> asLong();

}


short
Token::asShort() const
{
	PROC
	RETURN _value -> asShort();
}


bool
Token::asBool() const
{
	PROC
	RETURN _value -> asBool();

}


unsigned int
Token::asUInt() const
{
	PROC
	RETURN _value -> asUInt();

}

// new from EC
unsigned long
Token::asUlong() const
{
	PROC
	RETURN _value -> asUlong();

}


double
Token::asReal() const
{
	PROC
	RETURN _value -> asReal();

}


float
Token::asSingle() const
{
	PROC
	RETURN _value -> asShort();

}


char
Token::asCharacter() const
{
	PROC
	RETURN _value -> asCharacter();

}


string
Token::asString() const
{
	PROC
	RETURN _value -> asString();

}


const char *
Token::asCharPointer() const
{
	PROC
	RETURN _value -> asString().c_str();

}


const void *
Token::asPointer() const
{
	PROC
	RETURN _value -> asPointer();

}


Functor
Token::asFunctor() const
{
	PROC
	RETURN _value -> asFunctor();

}


PointerDescriptor
Token::asPointerDescriptor() const
{
	PROC
	RETURN _value -> asPointerDescriptor();

}


const Vec<Token> &
Token::asVector() const
{
	PROC
	RETURN _value -> asVector();

}


const vector<Token::BTree> &
Token::asBTreeVector() const
{
	PROC
	RETURN _value -> asBTreeVector();

}


const Token::BTree &
Token::asBTree() const
{
	PROC
	RETURN _value -> asBTree();
}


Token
Token::asToken() const
{
	PROC
	RETURN _value -> asToken();

}


Token
Token::operator<( const Token & token ) const
{

	PROC
	RETURN *_value < *token._value;

}

Token
Token::operator<=( const Token & token ) const
{

	PROC
	RETURN *_value <= *token._value;

}


Token
Token::operator>( const Token & token ) const
{

	PROC
	RETURN *_value > *token._value;

}


Token
Token::operator>=( const Token & token ) const
{

	PROC
	RETURN *_value >= *token._value;

}


Token
Token::operator==( const Token & token ) const
{

	PROC
	RETURN *_value == *token._value;

}


Token
Token::operator!=( const Token & token ) const
{

	PROC
	RETURN !(*_value == *token._value);

}


Token
Token::operator+( const Token &	token ) const
{

	PROC
	RETURN *_value + *token._value;

}


Token
Token::operator-( const Token &	token ) const
{
	PROC
	RETURN *_value - *token._value;

}


Token
Token::operator*( const Token &	token ) const
{
	PROC
	RETURN *_value * *token._value;

}


Token
Token::operator/( const Token &	token ) const
{
	PROC
	RETURN *_value / *token._value;

}


void
Token::printOn( ostream &	os ) const
{

	PROC
	os << "rawValue: " << rawValue() << " ID: " << int( id() );
	if( _value ) _value -> printOn( os );
	os << endl;
	ENDPROC

}


void
Token::setRawValue( const string & rawValue )
{
	PROC
	_rawValue = rawValue;
	ENDPROC
}


string
Token::rawValue() const
{
	PROC
	RETURN _rawValue;

}


void
Token::setDebug( const string & debug )
{
	PROC
	_debug = debug;
	ENDPROC
}


string
Token::debug() const
{
	PROC
	RETURN _debug;

}



PointerDescriptor::PointerDescriptor(
			Token::IDType	type,
			unsigned int	indirections,
			const void *	address )
		:
			_type( type ),
			_indirections( indirections ),
			_address( address )
{
	PROC
	ENDPROC
}


PointerDescriptor::PointerDescriptor( const PointerDescriptor & value )
	:
		_type( value._type ),
		_indirections( value._indirections ),
		_address( value._address )
{
	PROC
	ENDPROC
}


PointerDescriptor &
PointerDescriptor::operator=( const PointerDescriptor & value )
{
	PROC
	_type = value._type;
	_indirections = value._indirections;
	_address = value._address;
	RETURN *this;
}


Token::IDType
PointerDescriptor::type() const
{
	PROC
	RETURN _type;
}


unsigned int
PointerDescriptor::indirections() const
{
	PROC
	RETURN _indirections;
}


const void *
PointerDescriptor::raw() const
{
	PROC
	RETURN _address;
}


void
PointerDescriptor::printOn( ostream & os ) const
{
	PROC
	os << "_type: " << int(_type) << " _indirections: " << _indirections << " _address: " << _address << endl;
	ENDPROC
}


Functor::Functor(
		Type			type,
		const void * 		address,
		const Token &		returnType )
	:
		_type(type),
		_address( address ),
		_returnType( returnType )
{
	PROC
	ENDPROC
}


Functor::Functor( const Functor & value )
	:
		_type( value._type ),
		_address( value._address),
		_returnType( value._returnType )
{
	PROC
	ENDPROC
}


Functor&
Functor::operator=( const Functor & value )
{
	PROC
	_type = value._type;
	_address = value._address;
	_returnType = value._returnType;
	RETURN *this;
}


Functor::Type
Functor::type() const
{
	PROC
	RETURN _type;
}


const void *
Functor::address() const
{
	PROC
	RETURN _address;
}

Token
Functor::returnType() const
{
	PROC
	RETURN _returnType;
}


void
Functor::printOn( ostream & os ) const
{
	PROC
	os << "_type: " << int(_type) << " _address: " << _address << " _returnType: "; _returnType.printOn( os ); os << endl;
	ENDPROC
}


